#!/bin/bash
#
# System initialization script
#
# Mostly written by Lasse Collin.
# Contains code and comments written by Patrick Volkerding.
#
# Last modified: 2006-01-29

# Mount /proc and /sys as the very first step.
export PATH=/sbin:/bin
mount -n -t proc proc /proc
mount -n -t sysfs sysfs /sys

# Set the basic environment (including PATH) expected by boot scripts.
[[ ! $HAVE_RC_FUNCTIONS ]] && . /etc/rc.d/rc.functions --keep-environment

# Function to run emergency shell.
emergency() {
	echo
	echo 'You may now login and try to fix the problem.'
	echo 'The system will reboot when you exit the shell.'
	echo
	sync
	sulogin
	# Unmount just in case the admin forgot to do it.
	# FIXME: umount uses /etc/mtab so it may fail.
	echo 'Unmounting file systems.'
	umount -a -r
	mount -n -o remount,ro /
	sync
	echo 'Rebooting system.'
	sleep 1
	reboot -f
}


########
# udev #
########

# Mount /dev on tmpfs.
mount -n -t tmpfs -o "$MOUNT_DEV_FLAGS" tmpfs /dev

# Make sure that the required directories exist:
mkdir -pm 0755 /dev/pts /dev/shm

# Start udev.
( . /etc/rc.d/rc.udev start )


#############################
# Swap and root file system #
#############################

# Enable swap. Ignore unexisting devices and swap files for now; they are
# enabled later in this script.
swapon -a -e

# Check how throughly we should check the filesystems.
# Fastboot and forced full fsck are off by default.
FASTBOOT=no
FSCK_OPTIONS=

# Check if fastboot was requirested at shutdown:
[ -f /etc/fastboot ] && FASTBOOT=yes

# Check if forced filesystem check was requested at shutdown:
if [ -f /etc/forcefsck ]; then
	FASTBOOT=no
	FSCK_OPTIONS='-f'  # -f is supported by e2fsck and jfs_fsck.
fi

# Chech if fastboot=<on|off|forcefsck> has been specified on
# the kernel command line:
case "$fastboot" in
	on|yes)     FASTBOOT=yes ;;
	off|no)     FASTBOOT=no ;;
	forcefsck)  FASTBOOT=no; FSCK_OPTIONS="-f" ;;
esac

# Check that root is mounted read-only.
if touch /fsrwtest 2> /dev/null; then
	rm -f /fsrwtest
	sync
	echo
	echo 'ERROR: Root file system was already mounted read-write. Cannot'
	echo "       run file system checker (fsck). Add 'ro' as a kernel"
	echo "       argument, or use 'read-only' option in lilo.conf."
	echo
	sleep_verbose 30
else
	# The command line options and return codes are supported at least
	# by e2fsck, reiserfsck and jfs_fsck. fsck.xfs does nothing, thus
	# it is supported too. ;-)
	# NOTE: -a is deprecated; you should use -p instead. All the fsck
	# programs listed above support -p, but there are a few other
	# fsck programs left that support only -a. Since -a still works
	# and has identical behavior to -p, it seems better to use -a.
	RET=0
	if [[ $FASTBOOT = no ]]; then
		echo 'Checking the root file system:'
		# This calls the file system specific fsck and tries to
		# repair most errors automatically.
		fsck $FSCK_OPTIONS -T -C -a /
		RET=$?
	fi

	# RET:
	# 0 = All OK
	# 1 = Some errors were fixed
	# 2 = Some errors were fixed, reboot is required
	# 4 = Errors were left unfixed, admin needs to do something
	# 6 = Additional return code used by reiserfsck, similar to 4
	case $RET in
	0)
		;;
	1)
		echo
		echo 'Errors were found and fixed.'
		echo 'Continuing the boot process normally.'
		echo
		;;
	2)
		echo
		echo 'Root file system check found and corrected errors.'
		echo 'System will now reboot.'
		echo
		sync
		sleep_verbose 30
		reboot -f
		;;
	*)
		echo
		echo "Root file system check found errors (error code $RET)."
		echo 'Some of the errors were left unfixed.'
		emergency
		;;
	esac
	unset RET

	echo 'Remounting the root file system in read-write mode.'
	if ! mount -n -o remount,rw /; then
		echo
		echo 'Remounting the root file system in read-write mode failed.'
		emergency
	fi
fi


######################
# Root is read-write #
######################

# Set the system time from the hardware clock using hwclock --hctosys.
if type -P hwclock &> /dev/null; then
	# Check for a broken motherboard RTC clock (where ioports for rtc
	# are unknown) to prevent hwclock causing a hang:
	grep -wq rtc /proc/ioports && CLOCK_FLAGS= || CLOCK_FLAGS='--directisa'

	# This defaults to localtime if /etc/hardwareclock doesn't exist.
	if grep -wqs '^UTC' /etc/hardwareclock; then
		echo 'Setting system time from the hardware clock (UTC).'
		hwclock $CLOCK_FLAGS --hctosys --utc
	else
		echo 'Setting system time from the hardware clock (localtime).'
		hwclock $CLOCK_FLAGS --hctosys --localtime
	fi
fi

# Create a new /etc/mtab unless it is a symlink (usually to /proc/mounts).
if [ ! -L /etc/mtab ]; then
	# Remove old mtab and its lock files.
	rm -f /etc/mtab /etc/mtab~* /etc/mtab.tmp

	# Create a new /etc/mtab with a fake mount command. This way the
	# entry of the root file system will be on the very first line.
	mount -f -o remount,rw /

	# Add /proc and /sys to /etc/mtab. Don't put /dev to prevent
	# accidentally unmounting it.
	mount -f -t proc proc /proc
	mount -f -t sysfs sysfs /sys
	#mount -f -t tmpfs -o "$MOUNT_DEV_FLAGS" tmpfs /dev
fi

# Configure ISA Plug-and-Play devices:
if [ -f /etc/isapnp.conf ] && type -P isapnp &> /dev/null; then
	isapnp /etc/isapnp.conf
fi

# Load explicitly requested kernel modules. These might be required
# to use your ethernet card, sound card, or other optional hardware.
# Priority is given first to a script named "rc.modules.local", then
# to "rc.modules-$FULL_KERNEL_VERSION", and finally to the plain
# "rc.modules". Note that if /etc/rc.d/rc.modules.local is found, then
# "that will be the ONLY rc.modules script the machine will run, so
# "make sure it has everything in it that you need.
if [ -r /proc/modules ]; then
	for I in rc.modules.local rc.modules-$(< /proc/sys/kernel/osrelease) \
			rc.modules; do
		if [ -x "/etc/rc.d/$I" ]; then
			printf 'Running /etc/rc.d/%s' "$I"
			[ -L "$I" ] && printf ' -> %s' \
					"$(readlink /etc/rc.d/rc.modules)"
			printf ':\n'
			( . "/etc/rc.d/$I" start )
			break
		fi
	done
fi

# Configure runtime kernel parameters:
if [ -f /etc/sysctl.conf ] && type -P sysctl &> /dev/null; then
	sysctl -e -p /etc/sysctl.conf
fi

# Initialize the Logical Volume Manager.
# This won't start unless we find /etc/lvm/backup/ (LVM2). This is
# created by vgscan, so to use LVM you must run vgscan yourself
# the first time (and create some VGs and LVs).
if [ -d /etc/lvm/backup ]; then
	echo 'Initializing LVM (Logical Volume Manager):'

	# Check for device-mapper support in the kernel.
	HAS_DM=no
	if grep -wqs device-mapper /proc/devices; then
		# Try to load the device mapper module.
		modprobe dm-mod && HAS_DM=yes
	else
		HAS_DM=yes
	fi

	if [[ $HAS_DM = no ]]; then
		echo 'WARNING: /etc/lvm/backup exist but you have no device'
		echo '         mapper support in the kernel. Not enabling LVM.'
		sleep_verbose 30
	else
		# Scan for new volume groups.
		if vgscan --ignorelockingfailure 2> /dev/null; then
			# Make volume groups available to the kernel.
			# This should also make logical volumes available.
			vgchange -a y --ignorelockingfailure
		fi

		# Enable swapping again in case any LVs are used for swap.
		swapon -a -e
	fi
fi

# Check all the non-root filesystems:
if [[ $FASTBOOT = no ]]; then
	echo 'Checking non-root filesystems:'
	fsck $FORCEFSCK -C -R -A -a
fi

# Mount non-root file systems in fstab, but not NFS or SMB
# because TCP/IP is not yet configured, and not proc or sysfs
# because those have already been mounted.
echo 'Mounting non-root local filesystems:'
mount -a -v -t noproc,nosysfs,nonfs,nosmbfs,nocifs

# Enable swapping again. This is needed in case a swapfile is used
# or a swap partition is on LVM, as those can't be enabled until
# the filesystems they reside on have been mounted read-write.
# In contrast to earlier swapon commands in this script, the '-e'
# switch is omitted here, because this command shouldn't fail.
swapon -a


##############################
# Local file systems mounted #
##############################

# Attempt to umount and remove any leftover /initrd:
if [ -d /initrd ]; then
	umount /initrd/proc /initrd/sys /initrd 2> /dev/null
	rmdir /initrd 2> /dev/null
	blockdev --flushbufs /dev/ram0 2> /dev/null
fi

# Clean up temporary files from local disks.
rm -f /etc/nologin /etc/forcefsck /etc/fastboot /etc/dhcpc/*.pid

# Set the hostname. If no hostname is found, fall back on the old default.
[ ! -s /etc/HOSTNAME ] && echo 'darkstar.example.net' > /etc/HOSTNAME
hostname $(cut -f1 -d . /etc/HOSTNAME)

# Write the version of the running Linux kernel to /etc/motd.
if [[ $OVERWRITE_MOTD = yes ]] || [ ! -f /etc/motd ]; then
	rm -f /etc/motd
	uname -sr > /etc/motd
fi


################
# Entropy pool #
################

# Carry an entropy pool between reboots to improve randomness.
if [ -f /etc/random-seed ]; then
	echo 'Using /etc/random-seed to initialize /dev/urandom.'
	cat /etc/random-seed > /dev/urandom
fi

# Use the pool size from /proc, or 512 bytes.
if [ -r /proc/sys/kernel/random/poolsize ]; then
	POOL_SIZE=$(< /proc/sys/kernel/random/poolsize)
else
	POOL_SIZE=512
fi

# Make sure that entropy pool changes even between reboots even if
# shutdown fails.
dd if=/dev/urandom of=/etc/random-seed count=1 bs=$POOL_SIZE 2> /dev/null
unset POOL_SIZE

# Make sure that the permissions are strict.
chown root:root /etc/random-seed
chmod 0600 /etc/random-seed


############
# Services #
############

rc_services_start SYSINIT "${SYSINIT[@]}"


###################################
# /usr and /var must be available #
###################################

# Clean up temporary files. It's possible that these directories
# are not on local file system.
find /var/run ! -type d -print0 | xargs -0r rm -f --
rm -f /var/log/setup/tmp/* /var/state/saslauthd/saslauthd.pid \
		/var/lock/* /var/spool/uucp/LCK..* \
		/tmp/.X*lock /tmp/hunt* /tmp/core /core &> /dev/null

# Create a fresh utmp file:
: > /var/run/utmp
chown root:utmp /var/run/utmp
chmod 664 /var/run/utmp

# Create /tmp/.ICE-unix and /tmp/.X11-unix if they are not present.
if [ ! -e /tmp/.ICE-unix -o ! -e /tmp/.X11-unix ]; then
	mkdir -pm 1777 /tmp/.ICE-unix /tmp/.X11-unix
fi

# Save the contents of 'dmesg'.
dmesg -s 65536 > /var/log/dmesg

# Screen blanks after 15 minutes idle time, and powers down in one hour
# if the kernel supports APM or ACPI power management. Also silence the
# PC-speaker because most users don't like it:
setterm -blank 15 -powersave powerdown -powerdown 60 -blength 0

# End of rc.S
